import logging
from urllib.request import Request, urlopen

from .renderers import XMLRenderer
from .models import SendResponse, LabelResponse
from .parsers import XMLParser

logger = logging.getLogger(__name__)


def soap(tag):
    return '{http://www.w3.org/2003/05/soap-envelope}%s' % tag


class Client:
    def __init__(self, hostname, auth_info):
        self.base_url = 'http://{}/MRWEnvio.asmx'.format(hostname)
        self.auth_info = auth_info
        self.parser = XMLParser()
        self.renderer = XMLRenderer()

    def make_http_request(self, tag, request_data, extra_headers=None):
        # TODO: insert header 'SOAPAction: "http://www.mrw.es/GetEtiquetaEnvio"
        extra_headers = extra_headers or {}
        data = [
            {'soap:Header': self.auth_info.as_dict()},
            {'soap:Body': {
                tag: {
                    'mrw:request': request_data
                },
            }},
        ]
        namespaces = [
            ('soap', 'http://www.w3.org/2003/05/soap-envelope'),
            ('mrw', 'http://www.mrw.es/'),
        ]
        data = self.renderer.render({'soap:Envelope': data}, namespaces)
        logger.debug('Sending %s', data)
        headers = {'Content-type': 'application/soap+xml; charset=utf-8'}
        headers.update(extra_headers)
        return Request(self.base_url, data, headers, method='POST')

    def send(self, pickup_info, service_info):
        request_data = [pickup_info.as_dict(), service_info.as_dict()]
        req = self.make_http_request('mrw:TransmEnvio', request_data)
        with urlopen(req) as response:
            body = response.read()
            logger.debug('Received %s', body)

        response = self.parser.parse(body)
        message = response[soap('Envelope')][0][soap('Body')][0]
        response_obj = SendResponse.from_raw(message)
        return response_obj.result

    def get_label(self, get_label_request):
        request_data = [get_label_request.as_dict()[None]]
        req = self.make_http_request('mrw:GetEtiquetaEnvio', request_data)
        with urlopen(req) as response:
            body = response.read()
            logger.debug('Received %s', body)

        response = self.parser.parse(body)
        message = response[soap('Envelope')][0][soap('Body')][0]
        response_obj = LabelResponse.from_raw(message)
        return response_obj.result



from io import BytesIO
import xml.etree.ElementTree as etree


class ParseError(Exception):
    pass


class XMLParser:
    def parse(self, stream):
        parser = etree.XMLParser(encoding='utf-8')

        try:
            tree = etree.parse(BytesIO(stream), parser)
        except etree.ParseError as exc:
            raise ParseError(exc.msg)

        root = tree.getroot()
        data = self._xml_convert(root, unwrap=True)
        return {root.tag: data}

    def _xml_convert(self, element, unwrap=False):
        children = list(element)

        if len(children) > 0:
            data = []
            for child in children:
                data.append({child.tag: self._xml_convert(child, unwrap=True)})
            return {element.tag: data} if not unwrap else data

        # len(children) == 0
        return element.text



from unittest.mock import patch, MagicMock


def test_create_client():
    import pysagec
    url = '//user:pass@example.com/?franchise=12&subscriber=34&department=56'
    client = pysagec.create_client(url)
    assert isinstance(client, pysagec.Client)
    assert client.auth_info.username == 'user'
    assert client.auth_info.password == 'pass'
    assert client.auth_info.franchise_code == '12'
    assert client.auth_info.subscriber_code == '34'
    assert client.auth_info.departament_code == '56'


@patch('pysagec.Client', MagicMock())
def test_readme_example():
    import pysagec
    import datetime as dt

    url = '//user:pass@example.com/?franchise=12&subscriber=34&department=56'

    client = pysagec.create_client(url)

    pickup_info = pysagec.PickupInfo(
        pickup_address=pysagec.Address(street_name='Plaza de España',
                                       postal_code='36001',
                                       city='Pontevedra'),
        recipient_name='Juan Pérez',
        recipient_phone_number='555555555',
        comments='Por las mañanas.',
    )

    service_info = pysagec.ServiceInfo(
        number_of_packages=1,
        date=dt.date.today(),
        service_code='0000',
    )

    response = client.send(pickup_info, service_info)

    print(response.shipping_number)



from io import BytesIO
from xml.sax.saxutils import XMLGenerator


class RenderError(Exception):
    pass


class XMLRenderer:
    def render(self, data, namespaces=None):
        namespaces = namespaces or []
        stream = BytesIO()

        xml = XMLGenerator(stream, encoding='utf-8')
        xml.startDocument()

        if isinstance(data, (list, tuple)):
            raise RenderError('Renderer of iterables is not supported')

        keys = list(data.keys())
        if len(keys) != 1:
            raise RenderError('`data` should have only one key, the root')

        root_element = keys[0]

        attrs = {'xmlns:' + ns: url for ns, url in namespaces}
        xml.startElement(root_element, attrs)

        self._to_xml(xml, data[root_element])

        xml.endElement(root_element)
        xml.endDocument()
        return stream.getvalue()

    def _to_xml(self, xml, data):
        if isinstance(data, (list, tuple)):
            for item in data:
                self._to_xml(xml, item)

        elif isinstance(data, dict):
            for key, value in data.items():
                xml.startElement(key, {})
                self._to_xml(xml, value)
                xml.endElement(key)

        elif isinstance(data, bool):
            xml.characters('true' if data else 'false')

        elif data is None:
            # Don't output any value
            pass

        else:
            xml.characters(str(data))



import datetime as dt
import os
import subprocess
import sys

from pysagec import create_client
from pysagec import models
from pysagec import utils

import pytest


def launch(filename):
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])


@pytest.fixture
def pre_production_client():
    import logging
    logging.basicConfig(level=logging.DEBUG)

    url = os.environ.get(
        'TEST_URL',
        '//user:pass@example.com/?franchise=12&subscriber=34&department=56'
    )
    return create_client(url)


@pytest.mark.skipif(
    not os.environ.get('TEST_URL', False),
    reason='TEST_URL environment is no set',
)
def test_send(pre_production_client):
    address = models.Address()
    address.street_name = 'Plaza de España'
    address.postal_code = '36001'
    address.city = 'Pontevedra'
    pickup_info = models.PickupInfo()
    pickup_info.pickup_address = address
    pickup_info.recipient_name = 'Juan Pérez'
    service_info = models.ServiceInfo()
    service_info.number_of_packages = 1
    service_info.date = dt.date.today() + dt.timedelta(days=1)
    service_info.service_code = '0000'

    response = pre_production_client.send(pickup_info, service_info)
    print(repr(response))
    assert isinstance(response, models.SendResponseResult)
    assert hasattr(response, 'shipping_number')

    get_label = models.GetLabel()
    get_label.shipping_number = response.shipping_number

    response = pre_production_client.get_label(get_label)
    print(repr(response))
    assert isinstance(response, models.LabelResponseResult)
    assert hasattr(response, 'file')

    utils.base64_to_file(response.file, 'label.pdf')
    launch('label.pdf')



from pysagec import utils


def test_base64_to_file(tmpdir):
    data = b'AA=='
    filename = tmpdir.join('file.pdf')
    assert not filename.exists(), 'Temporary file exists'
    utils.base64_to_file(data, filename.strpath)
    assert filename.exists()


def test_get_hostname_from_url():
    assert 'example.com' == utils.get_hostname_from_url('//example.com/')



from .client import Client
from .models import Address, AuthInfo, Package, PickupInfo, ServiceInfo
from .utils import get_hostname_from_url


__all__ = [
    'Address', 'Client', 'Package', 'PickupInfo', 'ServiceInfo',
    'create_client'
]


def create_client(url):
    hostname = get_hostname_from_url(url)
    auth_info = AuthInfo.from_url(url)
    return Client(hostname, auth_info)



import pytest

from pysagec.renderers import XMLRenderer, RenderError


def test_xml_renderer_simple():
    renderer = XMLRenderer()
    stream = renderer.render({'root': None})
    assert b'<?xml version="1.0" encoding="utf-8"?>\n<root></root>' == stream


def test_xml_renderer_nested():
    renderer = XMLRenderer()
    data = {'root': {'children': [{'child1': 'Mike'}, {'child2': 'Liza'}]}}
    stream = renderer.render(data)
    assert (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root>'
        b'<children><child1>Mike</child1><child2>Liza</child2></children>'
        b'</root>'
    ) == stream


def test_xml_renderer_boolean():
    renderer = XMLRenderer()
    data = {'root': [{'yes': True}, {'no': False}]}
    stream = renderer.render(data)
    assert (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root><yes>true</yes><no>false</no></root>'
    ) == stream


def test_xml_render_raises_error_if_more_than_one_key_in_data():
    renderer = XMLRenderer()
    with pytest.raises(RenderError):
        renderer.render({'one': None, 'two': None})


def test_xml_render_raises_error_if_data_is_iterable():
    renderer = XMLRenderer()
    with pytest.raises(RenderError):
        renderer.render([{'one': None}, {'two': None}])


def test_xml_renderer_namespaces():
    renderer = XMLRenderer()
    namespaces = [('foo', 'http://example.com/ns')]
    stream = renderer.render({'root': None}, namespaces=namespaces)
    assert (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root xmlns:foo="http://example.com/ns"></root>'
    ) == stream



import base64
from urllib.parse import urlsplit


def base64_to_file(data, filename):
    data = base64.b64decode(data)
    with open(filename, 'wb') as fp:
        fp.write(data)


def get_hostname_from_url(url):
    url = urlsplit(url)
    return url.hostname



import pytest

from pysagec import parsers


def test_parser_simple():
    stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root></root>'
    )
    parser = parsers.XMLParser()
    data = parser.parse(stream)
    assert {'root': None} == data


def test_parser_simple_with_namespace():
    stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root xmlns="http://example.com"></root>'
    )
    parser = parsers.XMLParser()
    data = parser.parse(stream)
    assert {'{http://example.com}root': None} == data


def test_parser_one_children():
    stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root><tag>Value</tag></root>'
    )
    parser = parsers.XMLParser()
    data = parser.parse(stream)
    assert {'root': [{'tag': 'Value'}]} == data


def test_parser_one_children_with_two_namespaces():
    stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root'
        b' xmlns="http://example.com"'
        b' xmlns:ns="http://example.org">'
        b'<ns:tag>Value</ns:tag>'
        b'</root>'
    )
    parser = parsers.XMLParser()
    data = parser.parse(stream)
    assert ({
        '{http://example.com}root': [{'{http://example.org}tag': 'Value'}]
    }) == data


def test_parser_many_children():
    stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root><tag1>Value 1</tag1><tag2>Value 2</tag2></root>'
    )
    parser = parsers.XMLParser()
    data = parser.parse(stream)
    assert {'root': [{'tag1': 'Value 1'}, {'tag2': 'Value 2'}]} == data


def test_parser_nested():
    stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root><children>'
        b'<child1>Value 1</child1><child2>Value 2</child2>'
        b'</children></root>'
    )
    parser = parsers.XMLParser()
    data = parser.parse(stream)
    assert ({'root': [{
        'children': [{'child1': 'Value 1'}, {'child2': 'Value 2'}]}
    ]}) == data


def test_parser_error():
    bad_stream = (
        b'<?xml version="1.0" encoding="utf-8"?>\n'
        b'<root><tag1>Value 1</tag2>'
    )
    parser = parsers.XMLParser()
    with pytest.raises(parsers.ParseError):
        parser.parse(bad_stream)



from pysagec import base

import pytest


def test_field():
    f = base.Field('tag')
    assert f.__get__(None, None) is f


def test_string():
    f = base.String('tag')
    assert 'String' in repr(f)


def test_nested():
    f = base.Nested('tag', base.Model)
    assert 'Nested' in repr(f)


def test_model_as_dict():
    class MyModel(base.Model):
        root_tag = 'root'
        prop1 = base.String('tag1')
        prop2 = base.String('tag2')

    model = MyModel(prop1=42)
    model.prop2 = 'foo'

    assert model.prop1 == 42
    assert {'root': [{'tag1': 42}, {'tag2': 'foo'}]} == model.as_dict()


def test_model_default():
    class MyModel(base.Model):
        root_tag = 'root'
        prop = base.String('tag', default='x')

    model = MyModel()
    assert model.prop == 'x'


def test_field_ignore_if_none_do_not_show_field():
    class MyModel(base.Model):
        root_tag = 'root'
        prop1 = base.Field('tag1', default=None, ignore_if_none=False)
        prop2 = base.Field('tag2', default=None, ignore_if_none=True)

    model = MyModel()
    assert {'root': [{'tag1': None}]} == model.as_dict()


def test_nested_ignore_if_none_do_not_show_field():
    class ChildModel(base.Model):
        root_tag = None
        leaf1 = base.String('leaf1')
        leaf2 = base.String('leaf2')

    class ParentModel(base.Model):
        root_tag = 'root'
        prop = base.Nested('tag', ChildModel, unwrap=True, ignore_if_none=True)

    model = ParentModel(prop=None)
    expected = {'root': []}
    assert expected == model.as_dict()


def test_nested_single_unwrap():
    class ChildModel(base.Model):
        root_tag = None
        leaf1 = base.String('tag1')
        leaf2 = base.String('tag2')

    class ParentModel(base.Model):
        root_tag = 'root'
        prop = base.Nested('tag', ChildModel, unwrap=True)

    model = ParentModel()
    expected = {'root': [{'tag': [{'tag1': None}, {'tag2': None}]}]}
    assert expected == model.as_dict()


def test_nested_single_wrap():
    class ChildModel(base.Model):
        root_tag = 'child'
        leaf1 = base.String('leaf1')
        leaf2 = base.String('leaf2')

    class ParentModel(base.Model):
        root_tag = 'root'
        prop = base.Nested('tag', ChildModel, unwrap=False)

    model = ParentModel()
    expected_nested = {'child': [{'leaf1': None}, {'leaf2': None}]}
    expected = {'root': [{'tag': expected_nested}]}
    assert expected == model.as_dict()


def test_nested_many():
    class ChildModel(base.Model):
        root_tag = 'child'
        leaf1 = base.String('leaf1')
        leaf2 = base.String('leaf2')

    class ParentModel(base.Model):
        root_tag = 'root'
        prop = base.Nested('tag', ChildModel, many=True)

    model = ParentModel()
    expected_nested = [{'child': [{'leaf1': None}, {'leaf2': None}]}]
    expected = {'root': [{'tag': expected_nested}]}
    assert expected == model.as_dict()


def test_model_eq():
    class MyModel(base.Model):
        root_tag = 'root'
        prop = base.String('tag')

    m1 = MyModel(prop='42')
    m2 = MyModel(prop='42')
    m3 = MyModel(prop='2')
    assert m1 is not m2
    assert m1 == m2
    assert m1 != m3


def test_model_from_dict():
    class MyModel(base.Model):
        root_tag = 'root'
        prop = base.String('tag')

    data = {'root': [{'prop': '42'}]}
    model = MyModel.from_dict(data)

    expected_model = MyModel(prop='42')
    assert expected_model.prop == model.prop
    assert expected_model == model


def test_model_nested_from_dict():
    class ChildModel(base.Model):
        root_tag = 'child'
        leaf1 = base.String('leaf1')
        leaf2 = base.String('leaf2')

    class ParentModel(base.Model):
        root_tag = 'root'
        prop = base.Nested('tag', ChildModel, unwrap=False)

    nested = {'child': [{'leaf1': '42'}, {'leaf2': '3'}]}
    data = {'root': [{'tag': nested}]}
    model = ParentModel.from_dict(data)

    expected_model = ParentModel(tag=[ChildModel(leaf1='42', leaf2='3')])
    assert expected_model == model


def test_model_with_root_tag_none():
    class MyModel(base.Model):
        root_tag = None
        prop1 = base.String('tag1')
        prop2 = base.String('tag2')

    model = MyModel(prop1='1', prop2='two')
    assert {None: [{'tag1': '1'}, {'tag2': 'two'}]} == model.as_dict()


def test_model_from_raw():
    class MyModel(base.Model):
        root_tag = 'root'
        prop = base.String('tag')

    raw = {'root': [{'tag': '42'}, {'unknown_tag': 'x'}]}
    model = MyModel.from_raw(raw)

    expected_model = MyModel(prop='42')
    assert expected_model.prop == model.prop
    assert expected_model == model


def test_model_from_raw_if_bad_raw():
    class MyModel(base.Model):
        root_tag = 'root'
        prop = base.String('tag')

    raw = {'root': [{'tag': '42', 'bad_tag': 'xx'}]}
    with pytest.raises(ValueError):
        MyModel.from_raw(raw)


def test_model_from_raw_nested():
    class ChildModel(base.Model):
        root_tag = None
        leaf1 = base.String('tag1')
        leaf2 = base.String('tag2')

    class ParentModel(base.Model):
        root_tag = 'root'
        prop = base.Nested('tag', ChildModel, unwrap=True)

    raw = {'root': [{'tag': [{'tag1': '1'}, {'tag2': 'two'}]}]}
    model = ParentModel.from_raw(raw)

    expected_model = ParentModel(prop=ChildModel(leaf1='1', leaf2='two'))
    assert expected_model == model



from contextlib import contextmanager
from unittest.mock import patch, MagicMock

import pytest

from pysagec import Client


class MockModel:
    def as_dict(self):
        return {}


class MockLabelRequest:
    def as_dict(self):
        return {None: []}


@contextmanager
def patched_client(body_response):
    with patch('pysagec.client.urlopen', read=b'a') as urlopen_:
        response = urlopen_.return_value.__enter__.return_value
        response.read.return_value = body_response
        client = Client('example.com', MockModel())
        get_label_request = MagicMock()
        get_label_request.as_dict.return_value = {None: []}
        yield client

    assert urlopen_.called


def test_client_init():
    client = Client('example.com', MockModel())
    assert client.base_url == 'http://example.com/MRWEnvio.asmx'


def test_client_make_http_request():
    client = Client('example.com', MockModel())
    req = client.make_http_request('tag', {})
    assert req.get_method() == 'POST'


@pytest.mark.parametrize('body_response,status,shipping_number', [
    (
        (b'<?xml version="1.0" encoding="utf-8"?>'
         b'<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope"'
         b' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
         b' xmlns:xsd="http://www.w3.org/2001/XMLSchema">'
         b'<soap:Body>'
         b'<TransmEnvioResponse xmlns="http://www.mrw.es/">'
         b'<TransmEnvioResult>'
         b'<Estado>1</Estado>'
         b'<Mensaje>La fecha de recogida solicitada es 01/08/2016.</Mensaje>'
         b'<NumeroSolicitud>0330550767020160615223238261</NumeroSolicitud>'
         b'<NumeroEnvio>033050000050</NumeroEnvio>'
         b'<Url>http://sagec-test.mrw.es/Panel.aspx</Url>'
         b'</TransmEnvioResult>'
         b'</TransmEnvioResponse>'
         b'</soap:Body>'
         b'</soap:Envelope>'),
        '1',
        '033050000050',
    ),
])
def test_send(body_response, status, shipping_number):
    with patched_client(body_response) as client:
        response = client.send(MockModel(), MockModel())
    assert status == response.status
    assert shipping_number == response.shipping_number


@pytest.mark.parametrize('body_response,status,file', [
    (
        (b'<?xml version="1.0" encoding="utf-8"?>'
         b'<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope"'
         b' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
         b' xmlns:xsd="http://www.w3.org/2001/XMLSchema">'
         b'<soap:Body>'
         b'<GetEtiquetaEnvioResponse xmlns="http://www.mrw.es/">'
         b'<GetEtiquetaEnvioResult>'
         b'<Estado>1</Estado>'
         b'<Mensaje/>'
         b'<EtiquetaFile>AA==</EtiquetaFile>'
         b'</GetEtiquetaEnvioResult>'
         b'</GetEtiquetaEnvioResponse>'
         b'</soap:Body>'
         b'</soap:Envelope>'),
        '1',
        'AA==',
    ),
    (
        (b'<?xml version="1.0" encoding="utf-8"?>'
         b'<soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope"'
         b' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
         b' xmlns:xsd="http://www.w3.org/2001/XMLSchema">'
         b'<soap:Body>'
         b'<GetEtiquetaEnvioResponse xmlns="http://www.mrw.es/">'
         b'<GetEtiquetaEnvioResult>'
         b'<Estado>0</Estado>'
         b'<Mensaje/>'
         b'</GetEtiquetaEnvioResult>'
         b'</GetEtiquetaEnvioResponse>'
         b'</soap:Body>'
         b'</soap:Envelope>'),
        '0',
        None,
    ),
])
def test_get_label(body_response, status, file):
    with patched_client(body_response) as client:
        response = client.get_label(MockLabelRequest())
    assert status == response.status
    assert file == response.file



from urllib.parse import urlsplit, parse_qs

from .base import Model, Nested, String


def key_or_none(qs, key):
    iterable = qs.get(key, [None])
    return iterable[0]


class AuthInfo(Model):
    root_tag = 'mrw:AuthInfo'

    franchise_code = String('mrw:CodigoFranquicia')
    subscriber_code = String('mrw:CodigoAbonado')
    departament_code = String('mrw:CodigoDepartamento', ignore_if_none=True)
    username = String('mrw:UserName')
    password = String('mrw:Password')

    @classmethod
    def from_url(cls, url):
        url = urlsplit(url)
        qs = parse_qs(url.query)
        return cls(
            username=url.username,
            password=url.password,
            franchise_code=key_or_none(qs, 'franchise'),
            subscriber_code=key_or_none(qs, 'subscriber'),
            departament_code=key_or_none(qs, 'department')
        )


class Address(Model):
    root_tag = 'mrw:Direccion'

    street_type = String('mrw:CodigoTipoVia')
    street_name = String('mrw:Via')
    street_number = String('mrw:Numero')
    remaining_details = String('mrw:Resto')
    postal_code = String('mrw:CodigoPostal')
    city = String('mrw:Poblacion')


class PickupInfo(Model):
    root_tag = 'mrw:DatosEntrega'

    pickup_address = Nested('mrw:Direccion', Address, unwrap=True)
    vat_number = String('mrw:Nif')
    recipient_name = String('mrw:Nombre')
    recipient_phone_number = String('mrw:Telefono')
    contact_phone_number = String('mrw:Contacto')
    contact_name = String('mrw:ALaAtencionDe')
    comments = String('mrw:Observaciones')


class Package(Model):
    root_tag = 'mrw:BultoRequest'

    height = String('mrw:Alto')
    length = String('mrw:Largo')
    width = String('mrw:Ancho')
    dimension = String('mrw:Dimension')
    reference = String('mrw:Referencia')
    weight = String('mrw:Peso')


class ServiceInfo(Model):
    root_tag = 'mrw:DatosServicio'

    date = String('mrw:Fecha')
    customer_reference = String('mrw:Referencia')
    franchise_delivery = String('mrw:EnFranquicia', default='N')
    service_code = String('mrw:CodigoServicio')
    packages = Nested('mrw:Bultos', Package, many=True)
    number_of_packages = String('mrw:NumeroBultos')
    weight = String('mrw:Peso')
    delivery_on_saturday = String('mrw:EntregaSabado', default='N')
    delivery_before_830 = String('mrw:Entrega830', default='N')
    delivery_after_time = String('mrw:EntregaPartirDe')
    management = String('mrw:Gestion', default='N')
    return_back = String('mrw:Retorno', default='N')
    immediate_confirmation = String('mrw:ConfirmacionInmediata', default='N')
    reimbursement = String('mrw:Reembolso', default='N')


class GetLabel(Model):
    root_tag = None

    shipping_number = String('mrw:NumeroEnvio')
    delimiter = String('mrw:SeparadorNumerosEnvio', ignore_if_none=True)
    date_range_start = String('mrw:FechaInicioEnvio', ignore_if_none=True)
    date_range_end = String('mrw:FechaFinEnvio', ignore_if_none=True)
    label_type = String('mrw:TipoEtiquetaEnvio', default='0')
    top_margin = String('mrw:ReportTopMargin', default='1100')
    left_margin = String('mrw:ReportLeftMargin', default=650)


class SendResponseResult(Model):
    root_tag = None

    message = String('{http://www.mrw.es/}Mensaje')
    shipping_number = String('{http://www.mrw.es/}NumeroEnvio')
    request_number = String('{http://www.mrw.es/}NumeroSolicitud')
    status = String('{http://www.mrw.es/}Estado')
    url = String('{http://www.mrw.es/}Url')


class SendResponse(Model):
    root_tag = '{http://www.mrw.es/}TransmEnvioResponse'
    result = Nested(
        '{http://www.mrw.es/}TransmEnvioResult',
        SendResponseResult,
        unwrap=True
    )


class LabelResponseResult(Model):
    root_tag = None

    file = String('{http://www.mrw.es/}EtiquetaFile')
    message = String('{http://www.mrw.es/}Mensaje')
    status = String('{http://www.mrw.es/}Estado')


class LabelResponse(Model):
    root_tag = '{http://www.mrw.es/}GetEtiquetaEnvioResponse'
    result = Nested(
        '{http://www.mrw.es/}GetEtiquetaEnvioResult',
        LabelResponseResult,
        unwrap=True
    )



from collections import OrderedDict as odict


class Empty:
    pass


def filter_empty(iterable):
    return list(item for item in iterable if item is not Empty)


class Field:
    def __init__(self, tag_name, *, default=None, ignore_if_none=False):
        self.tag_name = tag_name
        self.value = None
        self.default = default
        self.ignore_if_none = ignore_if_none

        # This value will be set by metaclass
        self.name = None

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        else:
            return obj.__dict__.get(self.name, self.default)

    def __set__(self, obj, value):
        self.value = value
        obj.__dict__[self.name] = value

    def as_dict(self, obj):
        value = self.__get__(obj)
        if self.ignore_if_none and value is None:
            return Empty
        return {self.tag_name: value}

    def to_data(self, value):
        return value


class String(Field):
    def __repr__(self):
        return '<String name={!r}>'.format(self.name)


class Nested(Field):
    def __init__(self, tag_name, model, default=None,
                 unwrap=False, many=False, **kwargs):
        super().__init__(tag_name, default=default, **kwargs)

        assert not (many and unwrap), ('If `many` is `True `unwrap` must be '
                                       'set to `False`')

        self.model = model
        self.unwrap = unwrap
        self.many = many
        self.default = [model()] if many else model()

    def as_dict(self, obj):
        data = self.__get__(obj)
        if data is None:
            return Empty if self.ignore_if_none else None

        if self.many:
            return {
                self.tag_name: filter_empty([item.as_dict() for item in data])
            }

        value = data.as_dict()
        if self.unwrap:
            return {self.tag_name: value[self.model.root_tag]}

        return {self.tag_name: value}

    def to_data(self, value):
        assert self.many is False, '`many=True` not supported'
        raw_data = {self.model.root_tag: value}
        return self.model.from_raw(raw_data)

    def __repr__(self):
        return '<Nested name={!r} model={!r}>'.format(self.name, self.model)


class ModelMeta(type):
    def __new__(mcs, clsname, bases, clsdict):
        fields = []
        for name, value in clsdict.items():
            if isinstance(value, Field):
                value.name = name  # Init `Field.name`
                fields.append(value)
        clsdict['fields'] = fields
        return type.__new__(mcs, clsname, bases, clsdict)

    @classmethod
    def __prepare__(mcs, clsname, bases):
        # By returning an OrderedDict instead of a normal dictionary,
        # the resulting definition order is easily captured.
        # See Python Cookbook, Third Edition. Chapter 9
        return odict()


class Model(metaclass=ModelMeta):
    root_tag = 'root'

    @classmethod
    def from_dict(cls, data):
        kwargs = {}
        for el in data[cls.root_tag]:
            kwargs.update(el)
        return cls(**kwargs)

    @classmethod
    def from_raw(cls, raw):
        raw_fields = raw[cls.root_tag]
        tags = {field.tag_name: field for field in cls.fields}
        data = {}
        for raw_field in raw_fields:
            if len(raw_field) != 1:
                raise ValueError('Raw structure is not valid')
            tag = list(raw_field.keys())[0]
            try:
                field = tags[tag]
            except KeyError:
                continue
            value = raw_field[tag]
            data.update({field.name: field.to_data(value)})
        return cls(**data)

    def __init__(self, **kwargs):
        for name, value in kwargs.items():
            setattr(self, name, value)

    def as_dict(self):
        return {
            self.root_tag: filter_empty([f.as_dict(self) for f in self.fields])
        }

    def __eq__(self, other):
        return self.as_dict() == other.as_dict()



from pysagec import models


def test_models_are_instantiable():
    a = models.AuthInfo()
    assert isinstance(a, models.Model)


def test_create_auth_info_from_url():
    url = '//user:pass@example.com/?franchise=12&subscriber=34&department=56'
    auth_info = models.AuthInfo.from_url(url)
    assert isinstance(auth_info, models.AuthInfo)
    assert auth_info.username == 'user'
    assert auth_info.password == 'pass'
    assert auth_info.franchise_code == '12'
    assert auth_info.subscriber_code == '34'
    assert auth_info.departament_code == '56'
